from __future__ import annotations

import pathlib

# shutil is nicer, but doesn't work: https://bugs.python.org/issue20849
from functools import partial
from shutil import copytree as _copytree

import pytest

copytree = partial(_copytree, dirs_exist_ok=True)


@pytest.fixture
def datadir(tmp_path, request):
    """
    Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    # this gets the module name (e.g. /path/to/module-qc-analysis-tools/tests/test_cli.py)
    # and then gets the directory by removing the suffix (e.g. /path/to/module-qc-analysis-tools/tests/test_cli)
    test_dir = pathlib.Path(request.module.__file__).with_suffix("")

    if test_dir.is_dir():
        copytree(test_dir, str(tmp_path))

    return tmp_path



from __future__ import annotations

from construct import (
    Bytes,
    Computed,
    GreedyBytes,
    Pass,
    Struct,
    Switch,
)

from itksn.common import EnumStr
from itksn.pixels.common import pcb_manufacturer

triplet_assembly_site = EnumStr(
    Bytes(1),
    Genova=b"0",
    Barcelona=b"1",
    Oslo=b"2",
    Milano=b"3",
    LBNL=b"4",
)

batch_number = {
    0: "RD53A",
    1: "ITkpix_v1",
    2: "ITkpix_v2",
    3: "ITkpix_v2",
    4: "ITkpix_v2",
    5: "ITkpix_v2",
    6: "ITkpix_v2",
    7: "ITkpix_v2",
    8: "ITkpix_v2",
    9: "ITkpix_v2",
    10: "ITkpix_v2",
    11: "ITkpix_v2",
    12: "ITkpix_v2",
    13: "ITkpix_v2",
    14: "ITkpix_v2",
    15: "ITkpix_v2",
}

fe_chip = Struct(
    "number" / Bytes(7),
    "batch_number" / Computed(lambda ctx: (int(ctx.number) & 0xF0000) >> 16),  # type: ignore[arg-type,return-value]
    "batch" / Computed(lambda ctx: batch_number[ctx.batch_number]),  # type: ignore[arg-type,return-value]
    "wafer" / Computed(lambda ctx: (int(ctx.number) & 0x0FF00) >> 8),  # type: ignore[arg-type,return-value]
    "row" / Computed(lambda ctx: (int(ctx.number) & 0x000F0) >> 4),  # type: ignore[arg-type,return-value]
    "column" / Computed(lambda ctx: (int(ctx.number) & 0x0000F) >> 0),  # type: ignore[arg-type,return-value]
)

fe_chip_version = EnumStr(
    Bytes(1),
    RD53A=b"0",
    ITkpix_v1=b"1",
    ITkpix_v1p1=b"2",
    ITkpix_v2=b"3",
    No_chip=b"9",
)

fe_chip_version_pcb = EnumStr(
    Bytes(1),
    RD53A=b"0",
    Prototype_ITkpix_v1=b"1",
    Pre_production_OS_ITkpix_v1=b"2",
    Pre_production_IS_ITkpix_v1=b"3",
    Production_OS_ITkpix_v2=b"4",
    Production_IS_ITkpix_v2=b"5",
    No_chip=b"9",
)


sensor = Struct(
    "manufacturer"
    / EnumStr(
        Bytes(1),
        V1_ADVACAM=b"0",
        V2_HLL=b"1",
        V3_FBK_planar=b"2",
        V4_HPK=b"3",
        V5_LFoundry=b"4",
        V6_MICRON=b"5",
        V7_CNM=b"6",
        V8_FBK_3D=b"7",
        V9_SINTEF=b"8",
        Dummy=b"9",
    ),
    "sensor_type"
    / EnumStr(
        Bytes(1),
        RD53A_test_structure=b"0",
        Single=b"1",
        Halfmoon_preproduction_Double_MS=b"2",
        Quad=b"3",
        Planar_diode_test_structure_1=b"4",
        Strip_test_structure_2=b"5",
        Mini_sensor_test_structure_3=b"6",
        Interpixel_capacitance_test_structure_4=b"7",
        Biasing_test_structure_5=b"8",
        ThreeD_diode_test_structure_6=b"9",
    ),
    "number" / Bytes(5),
)

bare_module = Struct(
    "FE_chip_version" / fe_chip_version,
    "sensor_type"
    / EnumStr(
        Bytes(1),
        No_sensor=b"0",
        Market_survey_sensor_tile=b"1",
        L0_inner_pixel_3D_sensor_tile=b"2",
        L0_inner_pixel_planar_sensor_tile=b"3",
        L1_inner_pixel_quad_sensor_tile=b"4",
        Outer_pixel_quad_sensor_tile=b"5",
        Dummy_sensor_tile=b"9",
    ),
    "number" / Bytes(5),
)

pcb_loading_site = EnumStr(
    Bytes(1),
    outdated=b"0",
    Oslo=b"2",
    TOPRO=b"5",
    NORBIT=b"6",
    CERN=b"7",
    unloaded=b"9",
)

pcb_reception_site = EnumStr(
    Bytes(1),
    Genova=b"0",
    Barcelona=b"1",
    Oslo=b"2",
    Milano=b"3",
    Bergen=b"4",
)

pcb = Struct(
    "FE_chip_version" / fe_chip_version_pcb,
    "PCB_manufacturer" / pcb_manufacturer,
    "number" / Bytes(5),
)

pcb_triplets = Struct(
    "FE_chip_version" / fe_chip_version_pcb,
    "PCB_manufacturer" / pcb_manufacturer,
    "loading"
    / Switch(
        lambda ctx: ctx.PCB_manufacturer, {"Dummy": Bytes(1)}, default=pcb_loading_site
    ),  # esdape hatch for some dummy/digitals
    "reception"
    / Switch(
        lambda ctx: ctx.PCB_manufacturer,
        {"Dummy": Bytes(1)},
        default=pcb_reception_site,
    ),  # esdape hatch for some dummy/digitals
    "number" / Bytes(3),
)


module = Struct(
    "FE_chip_version" / fe_chip_version,
    "PCB_manufacturer"
    / Switch(
        lambda ctx: ctx.FE_chip_version, {"RD53A": pcb_manufacturer}, default=Pass
    ),
    "number" / GreedyBytes,
)
triplet_module = Struct(
    "FE_chip_version" / fe_chip_version,
    "assembly_site" / triplet_assembly_site,
    "not_used" / Bytes(1),
    "number" / Bytes(4),
)


module_carrier = Struct(
    "module_type"
    / EnumStr(
        Bytes(1),
        Quad_module_carrier=b"0",
        Cell_loaded_quad_module_bottom_cover=b"1",
        Linear_triplet_module_carrier=b"2",
        Ring_triplet_module_carrier=b"3",
    ),
    "module_version"
    / EnumStr(
        Bytes(1),
        Not_used=b"0",
        Triplet_v1p0=b"1",
        Quad_v2p1=b"2",
        Quad_v4p1=b"3",
        Quad_v4p2=b"4",
        Quad_v4p3=b"5",
    ),
    "manufacturer" / Bytes(1),
    "number" / Bytes(4),
)

yy_identifiers = {
    # pixel modules and subcomponents
    "FE_chip_wafer": ("FW", "PG"),
    "FE_chip": ("FC", "PG"),
    "Planar_sensor_wafer_100um_thickness": ("W6", "PI", "PG"),
    "Planar_sensor_wafer_150um_thickness": ("W7", "PG"),
    "ThreeD_sensor_wafer": ("W8", "PI", "PG"),
    "L0_inner_pixel_3D_sensor_wafer_25x100um": ("W0", "PI"),
    "L0_inner_pixel_3D_sensor_wafer_50x50um": ("W1", "PI"),
    "L1_inner_pixel_sensor_wafer_100um_thickness": ("W2", "PI"),
    "Outer_pixel_sensor_wafer_150um_thickness": ("W3", "PG"),
    "Half_size_planar_sensor_tile_100um_thickness": ("S6", "PG"),
    "Half_size_planar_sensor_tile_150um_thickness": ("S7", "PG"),
    "Full_size_planar_sensor_tile_100um_thickness": ("S8", "PI"),
    "Full_size_planar_sensor_tile_150um_thickness": ("S9", "PG"),
    "Half_size_3D_sensor_tile_25x100um": ("SG", "PG"),
    "Full_size_3D_sensor_tile_25x100um": ("SH", "PG"),
    "Half_size_3D_sensor_tile_50x50um": ("SI", "PG"),
    "Full_size_3D_sensor_tile_50x50um": ("SJ", "PG"),
    "L0_inner_pixel_3D_sensor_tile_25x100um": ("S0", "PI"),
    "L0_inner_pixel_3D_sensor_tile_50x50um": ("S1", "PI"),
    "L1_inner_pixel_quad_sensor_tile": ("S2", "PI"),
    "Outer_pixel_quad_sensor_tile": ("S3", "PG"),
    "Planar_Sensor_test_structure_100um_thickness": ("ST", "PI"),
    "Planar_Sensor_test_structure_150um_thickness": ("SU", "PG"),
    "ThreeD_Sensor_test_structure_25x100um": ("SV", "PI", "PG"),
    "ThreeD_Sensor_test_structure_50x50um": ("SW", "PI", "PG"),
    "Planar_Sensor_half_moon_100um_thickness": ("HT", "PI"),
    "Planar_Sensor_half_moon_150um_thickness": ("HU", "PG"),
    "ThreeD_Sensor_half_moon_25x100um": ("HV", "PI", "PG"),
    "ThreeD_Sensor_half_moon_50x50um": ("HW", "PI", "PG"),
    "Single_bare_module": ("B1", "PG"),
    "Dual_bare_module": ("B2", "PG"),
    "Quad_bare_module": ("B4", "PG"),
    "Digital_single_bare_module": ("BS", "PG"),
    "Digital_quad_bare_module": ("BQ", "PG"),
    "FourInch_bare_module_gel_pack": ("G4", "PG"),
    "SixInch_bare_module_gel_pack": ("G6", "PG"),
    "Triplet_L0_Stave_PCB": ("PT", "PI"),
    "Triplet_L0_R0_PCB": ("P0", "PI"),
    "Triplet_L0_R0p5_PCB": ("P5", "PI"),
    "Quad_PCB": ("PQ", "PG"),
    "Dual_PCB": ("PD", "PG"),
    "PCB_test_coupon": ("PC", "PG"),
    "OB_wirebond_protection_roof": ("WP", "P"),
    "Triplet_L0_stave_module": ("MS", "PI"),
    "Triplet_L0_Ring0_module": ("M0", "PI"),
    "Triplet_L0_Ring0p5_module": ("M5", "PI"),
    "L1_quad_module": ("M1", "PI"),
    "Outer_system_quad_module": ("M2", "PG"),
    "Dual_chip_module": ("R2", "PG"),
    "Single_chip_module": ("R0", "PG"),
    "Digital_triplet_L0_stave_module": ("R6", "PI"),
    "Digital_triplet_L0_ring0_module": ("R7", "PI"),
    "Digital_triplet_L0_ring0p5_module": ("R8", "PI"),
    "Digital_quad_module": ("R9", "PG"),
    "Digital_L1_quad_module": ("R", "PI"),
    "Dummy_triplet_L0_stave_module": ("RT", "PI"),
    "Dummy_triplet_L0_ring0_module": ("RU", "PI"),
    "Dummy_triplet_L0_ring0p5_module": ("RV", "PI"),
    "Dummy_quad_module": ("RQ", "PG"),
    "Dummy_L1_quad_module": ("RR", "PG"),
    "Module_carrier": ("MC", "PG"),
    # dummy for testing of GUIs/tutorials
    "Dummy_FE_chip_wafer": ("XW", "PG"),
    "Dummy_tutorial_FE_chip": ("XF", "PG"),
    "Dummy_sensor_wafer (with dummy tiles/test structures": ("XH", "PG"),
    "Dummy_tutorial_sensor_tile": ("XS", "PG"),
    "Dummy_sensor_test_structure": ("XT", "PG"),
    "Dummy_sensor_half_moon": ("XH", "PG"),
    "Dummy_single_bare_module": ("BT", "PG"),
    "Dummy_quad_bare_module": ("BR", "PG"),
    "Dummy_tutorial_bare_module": ("XB", "PG"),
    "Dummy_bare_module_gel_pack": ("XG", "PG"),
    "Dummy_tutorial_PCB": ("XP", "PG"),
    "Dummy_PCB_test_coupon": ("XD", "PG"),
    "Dummy_OB_wirebond_protection_roof": ("XE", "PB"),
    "Dummy_tutorial_module": ("XM", "PG"),
    "Dummy_module_carrier": ("XC", "PG"),
}



from __future__ import annotations

from construct import (
    Bytes,
    Struct,
)

from itksn.common import EnumStr
from itksn.pixels.common import pcb_manufacturer

local_supports_production_type = EnumStr(
    Bytes(1),
    Pre_production=b"0",
    Production=b"1",
    Prototype=b"2",
    All_types=b"9",
)

local_supports_flavor = EnumStr(
    Bytes(1),
    L0=b"0",
    L1=b"1",
    R01=b"2",
    R0p5=b"3",
    R1=b"4",
)

local_supports_longeron_flavor = EnumStr(
    Bytes(1),
    L0=b"0",
    L1=b"1",
    R01=b"2",
    R0p5=b"3",
    R1=b"4",
    A_side=b"5",
    C_side=b"6",
)

local_supports = Struct(
    "layer" / Bytes(1),
    "production_type" / local_supports_production_type,
    "number" / Bytes(5),
)

local_supports_is = Struct(
    "production_type" / local_supports_production_type,
    "type" / local_supports_flavor,
    "number" / Bytes(5),
)

local_supports_longeron = Struct(
    "production_type" / local_supports_production_type,
    "type" / local_supports_longeron_flavor,
    "number" / Bytes(5),
)

local_supports_ihr = Struct(
    "production_type" / local_supports_production_type,
    "type" / local_supports_flavor,
    "position"
    / EnumStr(
        Bytes(1),
        Standard=b"1",
        Last=b"2",
    ),
    "number" / Bytes(4),
)

loaded_local_supports_ob_module = Struct(
    "production_type" / local_supports_production_type,
    "PCB_manufacturer" / pcb_manufacturer,
    "number" / Bytes(5),
)

# FIXME: is the rest not defined?
local_supports_frame_box = Struct(
    "type" / EnumStr(Bytes(1), Longeron=b"1", HR=b"2"), "number" / Bytes(6)
)

yy_identifiers = {
    # local supports
    "IS_capillary": ("CP", "PI"),
    "IS_end_tube": ("ET", "PI"),
    "IS_cooling_tube": ("CA", "PI"),
    "IS_bare_local_support_stave": ("SS", "PI"),
    "IS_bare_local_support_ring": ("RS", "PI"),
    "IS_ring_local_support_assembly_loaded_ring": ("RL", "PI"),
    "IS_barrel_stave_assembly_loaded_stave": ("SL", "PI"),
    "OB_Base_Block": ("BB", "PB"),
    "OB_Cooling_Block": ("CB", "PB"),
    "OB_TPG_Tile": ("GT", "PB"),
    "OB_Local_Support_Inserts": ("IN", "PB"),
    "OB_Gusset": ("RG", "PB"),
    "OB_Truss": ("LT", "PB"),
    "OB_Half_Ring_Shell": ("RS", "PB"),
    "OB_End_of_longeron_Bracket_End_Gusset": ("EG", "PB"),
    "OB_Pipe_Support": ("PS", "PB"),
    "OB_Evaporator_Sleeves": ("ES", "PB"),
    "OB_Cooling_Pipe_IHR": ("RP", "PB"),
    "OB_Cooling_Pipe_Longeron": ("LP", "PB"),
    "OB_Functional_Pipe_for_IHR": ("RE", "PB"),
    "OB_Functional_Pipe_for_Longeron": ("LE", "PB"),
    "OB_End_of_longeron_Support": ("EL", "PB"),
    "OB_Bare_Module_Cell": ("BC", "PB"),
    "OB_Functional_IHR": ("FR", "PB"),
    "OB_Functional_Longeron": ("FL", "PB"),
    "OB_Loaded_Module_Cell": ("LC", "PB"),
    "OB_Loaded_IHR": ("LR", "PB"),
    "OB_Loaded_Longeron": ("LL", "PB"),
    "OB_IHR_Handling_Frame": ("HR", "PB"),
    "OB_Longeron_Handling_Frame": ("HL", "PB"),
    "OB_Bare_Cell_Transport_Box": ("TB", "PB"),
    "OEC_Trapezoids": ("TZ", "PE"),
    "OEC_Electrical break": ("EB", "PE"),
    "OEC_Inner_Rim_Insert": ("II", "PE"),
    "OEC_Outer_Rim_mounting_lugs": ("ML", "PE"),
    "OEC_Inner_Rim_Closeout": ("IC", "PE"),
    "OEC_Outer_Rim_Closeout": ("OC", "PE"),
    "OEC_Pipe_Closeout_support_closeout": ("SC", "PE"),
    "OEC_Half_Sandwich": ("HS", "PE"),
    "OEC_Evaporator": ("EV", "PE"),
    "OEC_Bare_half_ring_assembly_Bare_support": ("BH", "PE"),
    "OEC_Loaded_local_support_loaded_support": ("LS", "PE"),
    "OEC_Handling_frame_support_frame": ("SF", "PE"),
    "OEC_Transport/storage_box_support_box": ("SB", "PE"),
    "Local_support_handling_frame_box": ("LB", "PE", "PI", "PB"),
    "High_voltage_group": ("VG", "PE", "PI", "PB"),
    "Serial_powering_scheme": ("SP", "PE", "PI", "PB"),
    # dummy local supports
    "Dummy_IS_capillary": ("YA", "PI"),
    "Dummy_IS_end_tube": ("YB", "PI"),
    "Dummy_IS_cooling_tube": ("YD", "PI"),
    "Dummy_IS_bare_local_support_stave": ("YE", "PI"),
    "Dummy_IS_bare_local_support_ring": ("YF", "PI"),
    "Dummy_IS_ring_local_support_assembly_loaded_ring": ("YG", "PI"),
    "Dummy_IS_barrel_stave_assembly_loaded_stave": ("YH", "PI"),
    "Dummy_OB_Base_Block": ("ZA", "PB"),
    "Dummy_OB_Cooling_Block": ("ZB", "PB"),
    "Dummy_OB_TPG_Tile": ("ZC", "PB"),
    "Dummy_OB_Local_Support_Inserts": ("ZD", "PB"),
    "Dummy_OB_Gusset": ("ZE", "PB"),
    "Dummy_OB_Truss": ("ZF", "PB"),
    "Dummy_OB_Half_Ring_Shell": ("ZG", "PB"),
    "Dummy_OB_End_of_longeron_Bracket_End_Gusset": ("ZH", "PB"),
    "Dummy_OB_Pipe_Support": ("ZI", "PB"),
    "Dummy_OB_Evaporator_Sleeves": ("ZJ", "PB"),
    "Dummy_OB_Cooling_Pipe_IHR": ("ZK", "PB"),
    "Dummy_OB_Cooling_Pipe_Longeron": ("ZL", "PB"),
    "Dummy_OB_Functional_Pipe_for_IHR": ("ZM", "PB"),
    "Dummy_OB_Functional_Pipe_for_Longeron": ("ZN", "PB"),
    "Dummy_OB_End_of_longeron_Support": ("ZO", "PB"),
    "Dummy_OB_Bare_Module_Cell": ("ZP", "PB"),
    "Dummy_OB_Functional_IHR": ("ZQ", "PB"),
    "Dummy_OB_Functional_Longeron": ("ZR", "PB"),
    "Dummy_OB_Loaded_Module_Cell": ("ZS", "PB"),
    "Dummy_OB_Loaded_IHR": ("ZT", "PB"),
    "Dummy_OB_Loaded_Longeron": ("ZU", "PB"),
    "Dummy_OB_IHR_Handling_Frame": ("ZV", "PB"),
    "Dummy_OB_Longeron_Handling_Frame": ("ZW", "PB"),
    "Dummy_OB_Bare_Cell_Transport_Box": ("ZX", "PB"),
    "Dummy_OEC_Trapezoids": ("YI", "PE"),
    "Dummy_OEC_Electrical break": ("YJ", "PE"),
    "Dummy_OEC_Inner_Rim_Insert": ("YK", "PE"),
    "Dummy_OEC_Outer_Rim_mounting_lugs": ("YL", "PE"),
    "Dummy_OEC_Inner_Rim_Closeout": ("YM", "PE"),
    "Dummy_OEC_Outer_Rim_Closeout": ("YN", "PE"),
    "Dummy_OEC_Pipe_Closeout_support_closeout": ("YO", "PE"),
    "Dummy_OEC_Half_Sandwich": ("YP", "PE"),
    "Dummy_OEC_Evaporator": ("YQ", "PE"),
    "Dummy_OEC_Bare_half_ring_assembly_Bare_support": ("YR", "PE"),
    "Dummy_OEC_Loaded_local_support_loaded_support": ("YS", "PE"),
    "Dummy_OEC_Handling_frame_support_frame": ("YT", "PE"),
    "Dummy_OEC_Transport/storage_box_support_box": ("YU", "PE"),
    "Dummy_Local_support_handling_frame_box": ("YV", "PE", "PI", "PB"),
    "Dummy_High_voltage_group": ("YW", "PE", "PI", "PB"),
    "Dummy_Serial_powering_scheme": ("YX", "PE", "PI", "PB"),
}



from __future__ import annotations

from typing import TypeVar

from construct import (
    Construct,
    Error,
    Switch,
)

# from construct-typing
ParsedType_co = TypeVar("ParsedType_co", covariant=True)  # pylint:disable=invalid-name
BuildTypes_contra = TypeVar("BuildTypes_contra", contravariant=True)


def subproject_switch(
    pg: Construct[ParsedType_co, BuildTypes_contra] | None = None,
    pi: Construct[ParsedType_co, BuildTypes_contra] | None = None,
    pe: Construct[ParsedType_co, BuildTypes_contra] | None = None,
    pb: Construct[ParsedType_co, BuildTypes_contra] | None = None,
) -> Construct[ParsedType_co, BuildTypes_contra]:
    """
    helper utility to pick up a different Construct for different subprojects
    """
    return Switch(
        lambda ctx: ctx.subproject_code,
        {
            "inner_pixel": pi or Error,
            "outer_pixel_barrel": pb or Error,
            "pixel_general": pg or Error,
            "pixel_endcaps": pe or Error,
        },
    )



from __future__ import annotations

from construct import (
    Bytes,
    Computed,
    Const,
    Pointer,
    Select,
    Struct,
    Switch,
)

from itksn.common import EnumStr

orientation = EnumStr(
    Bytes(1),
    normal=b"0",
    mirror=b"1",
)

optoboard = Struct(
    "production_version"
    / Select(
        Const(b"2"), Const(b"3"), Const(b"4"), Const(b"5")
    ),  # FIXME: "2", "3", "5" not defined in table 6
    "LpGBT_count"
    / EnumStr(
        Bytes(1),
        one=b"1",
        two=b"2",
        four=b"4",
    ),
    "number" / Bytes(5),
)

termination_board = Struct(
    "production_version" / Const(b"0"),
    "flavor"
    / EnumStr(
        Bytes(1),
        Normal_L_long=b"0",
        Normal_L_short=b"1",
        Mirror_L_long=b"2",
        Mirror_L_short=b"3",
        Normal_slim=b"4",
        Mirror_slim=b"5",
        Normal_extended_slim=b"6",  # FIXME: OB Type-1 Termination Board does not have this
    ),
    "number" / Bytes(5),
)

optobox_powerboard_connector = Struct(
    "production_version"
    / Select(
        Const(b"0"), Const(b"2"), Const(b"3")
    ),  # FIXME: "0" and "3" not in table 9
    "orientation" / orientation,
    "number" / Bytes(5),
)

optobox = Struct(
    "production_version" / Const(b"0"),
    "orientation" / orientation,
    "number" / Bytes(5),
)

canbus = Struct(
    "production_version" / Const(b"0"),
    "connector_type"
    / EnumStr(
        Bytes(1),
        six=b"6",
        eight=b"8",
    ),
    "number" / Bytes(5),
)

type0_component: dict[str, dict[tuple[str, str], str]] = {
    "Power_pigtail": {  # "PP"
        ("Barrel_Triplet", "NA"): "L0 Barrel Power Flex",
        ("Barrel_Quad", "F1"): "L1 Barrel Power Flex F1",
        ("Barrel_Quad", "F2"): "L1 Barrel Power Flex F2",
        ("Endcap_Mixed", "F1"): "Triplet Power T-Rex",
        ("Endcap_Mixed", "F2"): "Triplet Power Jumper",
    },
    "Rigid_flex": {  # "RF"
        ("Endcap_Mixed", "NA"): "Coupled Ring",
        ("Endcap_Quad", "NA"): "Quad Ring",
        ("Endcap_Ring05_Triplet", "NA"): "Intermediate Ring",
    },
    "Pigtail": {  # "PG"
        ("Barrel_Triplet", "NA"): "L0 Barrel Data Flex",
        ("Barrel_Quad", "F1"): "L1 Barrel Data Flex F1",
        ("Barrel_Quad", "F2"): "L1 Barrel Data Flex F2",
        ("Barrel_Quad", "F3"): "L1 Barrel Data Flex F3",
        ("Barrel_Quad", "F4"): "L1 Barrel Data Flex F4",
        ("Endcap_Quad", "NA"): "Quad Module Z-Ray",
        ("Endcap_Ring0_Triplet", "F1"): "R0 Data Flex F1",
        ("Endcap_Ring0_Triplet", "F2"): "R0 Data Flex F2",
        ("Endcap_Ring0_Triplet", "F3"): "R0 Data Flex F3",
        ("Endcap_Ring05_Triplet", "F1"): "R0.5 Data Flex F1",
        ("Endcap_Ring05_Triplet", "F2"): "R0.5 Data Flex F2",
        ("Endcap_Mixed", "F1"): "Type-0 to PP0 F1",
        ("Endcap_Mixed", "F2"): "Type-0 to PP0 F2",
    },
}

is_type0_cable = Struct(
    "stage"
    / EnumStr(
        Bytes(1),
        Pre_production=b"0",
        Production=b"1",
        Dummy=b"9",
    ),
    "region"
    / EnumStr(
        Bytes(1),
        Barrel_Triplet=b"0",
        Barrel_Quad=b"1",
        Endcap_Ring0_Triplet=b"2",
        Endcap_Ring05_Triplet=b"3",
        Endcap_Quad=b"4",
        Endcap_Mixed=b"5",
    ),
    "flavor"
    / EnumStr(
        Bytes(1),
        NA=b"0",
        F1=b"1",
        F2=b"2",
        F3=b"3",
        F4=b"4",
    ),
    "component"
    / Computed(
        lambda ctx: type0_component[ctx._.component_code][(ctx.region, ctx.flavor)],  # type: ignore[arg-type,return-value]
    ),
    "number" / Bytes(4),
)

type1_production_version = EnumStr(
    Bytes(1),
    Prototype=b"0",
    Pre_production=b"1",
    Production=b"2",
    Dummy=b"9",
)

is_type1_cable = Struct(
    "production_version" / type1_production_version,
    "flavor"
    / EnumStr(
        Bytes(1),
        L0L1=b"0",
        L02xL1=b"1",
        Coupled_Ring=b"2",
        Intermediate_Ring=b"3",
        QR1=b"4",
        QR2=b"5",
    ),
    "number" / Bytes(5),
)

pp1 = Struct(
    "production_version" / type1_production_version,
    "glenair_part"
    / EnumStr(
        Bytes(1),
        Type_2_Header=b"0",
        Type_1_Receptacle=b"1",
        Strain_Arm=b"3",
        Filter_Plug=b"9",
    ),
    "flavor"
    / EnumStr(
        Bytes(1),
        N=b"0",
        A=b"1",
        B=b"2",
        C=b"3",
        D=b"4",
        E=b"5",
    ),
    "number" / Bytes(4),
)

pi_type0_pp0 = Struct(
    "production_version"
    / EnumStr(
        Bytes(1),
        Prototype=b"0",
        Pre_production=b"1",
        Production=b"2",
        Dummy=b"9",
    ),
    "flavor"
    / EnumStr(
        Bytes(1),
        Zp_3Stave=b"0",
        Zm_3Stave=b"1",
        Zp_2Stave=b"2",
        Zm_2Stave=b"3",
        Coupled_Ring=b"4",
        Intermediate_Ring=b"5",
        Quad_Ring=b"6",
    ),
    "number" / Bytes(5),
)

# PB-PL, PB-PG
pb_type0_cable = Struct(
    "type" / EnumStr(Bytes(1), Flat=b"0", Inclined=b"1", Inclined_test_coupon=b"2"),
    "flavor"
    / Switch(
        lambda ctx: ctx.type,
        {
            "Flat": EnumStr(Bytes(1), Bottom=b"0", Top=b"1"),
            "Inclined": EnumStr(
                Bytes(1),
                Front=b"0",
                Back=b"1",
                Front_Last_ring=b"2",
                Back_Last_ring=b"3",
            ),
        },
    ),
    "version"
    / EnumStr(
        Bytes(1),
        Pre_production=b"0",
        Production=b"1",
        Prototype=b"9",
    ),
    "number" / Bytes(4),
)

# PB-RF
pb_type0_pp0 = Struct(
    "type" / EnumStr(Bytes(1), Flat=b"0", Inclined=b"1", Inclined_test_coupon=b"2"),
    "flavor"
    / Switch(
        lambda ctx: ctx.type,
        {
            "Flat": EnumStr(
                Bytes(1), Short_L2=b"0", Long_L2=b"1", Short_L3_L4=b"2", Long_L3_L4=b"3"
            ),
            "Inclined": EnumStr(
                Bytes(1),
                L2_SP1=b"0",
                L2_SP2=b"1",
                L3_SP1=b"2",
                L3_SP2=b"3",
                L4_SP1=b"4",
                L4_SP2=b"5",
                Dummy=b"9",
            ),
            "Inclined_test_coupon": EnumStr(  # FIXME: not specified in Table 21
                Bytes(1),
                L2_SP1=b"0",
                L2_SP2=b"1",
                L3_SP1=b"2",
                L3_SP2=b"3",
                L4_SP1=b"4",
                L4_SP2=b"5",
                Dummy=b"9",
            ),
        },
    ),
    "version"
    / EnumStr(
        Bytes(1),
        Pre_production=b"0",
        Production=b"1",
        Prototype=b"9",
    ),
    "number" / Bytes(4),
)

pb_type1_power = Struct(
    "type"
    / EnumStr(
        Bytes(1),
        Flat=b"0",
        Inclined=b"1",
    ),
    "serial_power_chains"
    / EnumStr(
        Bytes(1),
        F1=b"4",
        F2=b"2",
    ),
    "number" / Bytes(5),
)

pb_type1_data_bundle = Struct(
    "flavor"
    / EnumStr(
        Bytes(2),
        Flat_L2_Normal_Slim=b"00",
        Flat_L2_Mirror_Slim=b"01",
        Flat_L3_L4_Mirror_Slim=b"02",
        Flat_L3_L4_Normal_Slim=b"03",
        Inclined_L2_Normal_L_short=b"04",
        Inclined_L2_Normal_L_long=b"05",
        Inclined_L2_Mirror_L_short=b"06",
        Inclined_L2_Mirror_L_long=b"07",
        Inclined_L3_Normal_Slim10=b"08",
        Inclined_L3_Mirror_Slim10=b"09",
        Inclined_L4_Mirror_Slim16=b"10",
        Inclined_L4_Normal_Slim=b"11",
        Inclined_L3_L4_Mirror_Slim12=b"12",
        Inclined_L3_L4_Normal_Slim12=b"13",
    ),
)

pb_type1_data_inclined = Struct(
    "version" / Bytes(1),
    "manufacturer" / Bytes(1),
)

pb_type1_data = Struct(
    "_reserved" / Pointer(2, Bytes(2)),
    "data"
    / Switch(
        lambda ctx: ctx._reserved,  # pylint: disable=protected-access
        {
            b"00": pb_type1_data_inclined,
        },
        default=pb_type1_data_bundle,
    ),
    "length" / Bytes(2),
    "number" / Bytes(3),
)

pe_type0_data_mapping = Switch(
    lambda ctx: ctx.layer,
    {
        "L2": EnumStr(
            Bytes(1),
            ring15_Front=b"0",
            ring15_Back=b"1",
            ring611_Front=b"2",
            ring611_Back=b"3",
        ),
        "L3": EnumStr(
            Bytes(1),
            Front_5downlinks=b"0",
            Back_5downlinks=b"1",
            Front_6downlinks=b"2",
            Back_6downlinks=b"3",
        ),
        "L4": EnumStr(
            Bytes(1),
            ring17_Front_4downlinks=b"0",
            ring17_Back_4downlinks=b"1",
            ring17_Front_5downlinks=b"2",
            ring17_Back_5downlinks=b"3",
            ring89_Front_4downlinks=b"4",
            ring89_Back_4downlinks=b"5",
            ring89_Front_5downlinks=b"6",
            ring89_Back_5downlinks=b"7",
        ),
    },
)

pe_type0_data = Struct(
    "layer" / EnumStr(Bytes(1), L2=b"2", L3=b"3", L4=b"4", All=b"9"),
    "flavor" / pe_type0_data_mapping,
    "reserved" / Const(b"0"),
    "number" / Bytes(4),
)

pe_type0_power = Struct(
    "layer" / EnumStr(Bytes(1), L2=b"2", L3=b"3", L4=b"4", All=b"9"),
    "flavor" / Bytes(1),
    "reserved" / Const(b"0"),
    "number" / Bytes(4),
)

# FIXME: merge with pe_type0_power?
# 1P (PP1_connector) or PB (Power_bustape)
pe_type1 = Struct(
    "flavor" / Bytes(1),
    "reserved" / Const(b"0"),
    "length"
    / EnumStr(
        Bytes(1),
        _50cm=b"0",
        _250cm=b"9",
    ),
    "number" / Bytes(4),
)

type2 = Struct(
    "flavor"
    / Switch(
        lambda ctx: ctx._.component_code,
        {
            "Type_2_power_cable": EnumStr(Bytes(1), normal=b"1", abnormal=b"2"),
            "Type_2_optobox_cable": Const(b"0"),
            "PP2_box": EnumStr(
                Bytes(1),
                F1=b"1",
                F2=b"2",
                F3=b"3",
                F4=b"4",
                F5=b"5",
                F6=b"6",
                F7=b"7",
                F8=b"8",
            ),
        },
    ),
    "number" / Bytes(6),
)

type3 = Struct(
    "flavor" / Const(b"0"),
    "numbers" / Bytes(6),
)

type4 = Struct(
    "flavor"
    / EnumStr(
        Bytes(1),
        Pre_production_non_rad_hard=b"0",
        Pre_production_rad_hard=b"1",
        Production=b"2",
    )
)

mops_chip = Struct(
    "reserved" / Const(b"00"),
    "production_version"
    / EnumStr(
        Bytes(1),
        Pre_production=b"0",
        Production=b"3",
    ),
    "vendor" / Bytes(4),
)

yy_identifiers = {
    # services
    "Optoboard": ("OB", "PG"),
    "Optoboard_termination_board": ("OT", "PG"),  # codespell:ignore
    "Optobox": ("OX", "PG"),
    "Optobox_powerbox": ("OW", "PG"),
    "Optobox_connector_board": ("OC", "PG"),
    "Optobox_optical_fan_out": ("OF", "PG"),
    "Optopanel": ("OS", "PG"),
    "Optopanel_cooling_plate": ("OO", "PG"),
    "Optobox_powerboard": ("OP", "PG"),
    "LpGBT_chip": ("OL", "PG"),
    "GBCR_chip": ("OG", "PG"),
    "Vtrx_module": ("OV", "PG"),
    "Bpol2V5_chip": ("O5", "PG"),
    "Bpol2V5_carrier_board": ("OK", "PG"),
    "Bpol12V_chip": ("O2", "PG"),
    "MOPS_chip": ("MP", "PG"),
    "Power_cables": ("OI", "PG"),
    "CAN_bus_cable": ("OD", "PG"),
    "Pigtail": ("PG", "PI", "PB"),
    "Rigid_flex": ("RF", "PI", "PB"),
    "Data_PP0": ("DP", "PI", "PE"),
    "Power_pigtail": ("PP", "PI", "PE"),
    "Power_bustape": ("PB", "PI", "PB", "PE"),
    "Bare_bustape": ("NB", "PE"),
    "Pigtail_panel": ("PL", "PB"),
    "PP0": ("0P", "PI"),  # FIXME: conflict with Triplet_L0_R0_PC
    "Finger": ("FI", "PI"),
    "Type_1_Data_link ": ("D1", "PI", "PB", "PE"),
    "Type_1_Power_DCS_line": ("P1", "PI", "PB", "PE"),
    "Environmental_link": ("E1", "PI", "PB", "PE"),
    "PP1_connector": ("1P", "PI", "PB", "PE"),
    "PP1_connector_pieces_segments": ("CS", "PI", "PB", "PE"),
    "strain_relief": ("SR", "PB"),
    "Type_2_power_cable": ("P2", "PI", "PB", "PE"),
    "Type_2_optobox_cable": ("20", "PI", "PB", "PE"),
    "PP2_box": ("2P", "PG"),
    "Type_3_HV_cable": ("P3", "PI", "PB", "PE"),
    "Type_3_LV_cable": ("L3", "PI", "PB", "PE"),
    "MOPS_cable": ("M3", "PG"),
    "TiLock": ("TL", "PG"),
    "OPTO_MOPS": ("OM", "PG"),
    "PP3_power": ("3P", "PG"),
    # Dummy services
    "Dummy_Optoboard": ("Q1", "PG"),
    "Dummy_Optoboard_termination_board": ("Q2", "PG"),  # codespell:ignore
    "Dummy_Optobox": ("Q3", "PG"),
    "Dummy_Optobox_powerbox": ("Q4", "PG"),
    "Dummy_Optobox_connector_board": ("Q5", "PG"),
    "Dummy_Optobox_optical_fan_out": ("Q6", "PG"),
    "Dummy_Optopanel": ("Q7", "PG"),
    "Dummy_Optopanel_cooling_plate": ("Q8", "PG"),
    "Dummy_Optobox_powerboard": ("Q9", "PG"),
    # "Dummy_LpGBT_chip": ("", "PG"), # FIXME: missing dummy
    "Dummy_GBCR_chip": ("QA", "PG"),
    "Dummy_Vtrx_module": ("QB", "PG"),
    "Dummy_Bpol2V5_chip": ("QC", "PG"),
    "Dummy_Bpol2V5_carrier_board": ("QD", "PG"),
    "Dummy_Bpol12V_chip": ("QE", "PG"),
    "Dummy_MOPS_chip": ("QF", "PG"),
    "Dummy_Power_cables": ("QG", "PG"),
    "Dummy_CAN_bus_cable": ("QH", "PG"),
    "Dummy_Pigtail": ("QI", "PB"),
    "Dummy_Rigid_flex": ("QK", "PI", "PB"),  # FIXME: "QJ" skipped
    "Dummy_Data_PP0": ("QL", "PI", "PE"),  # FIXME: typo in table 3
    "Dummy_Power_pigtail": ("QM", "PI", "PE"),
    "Dummy_Power_bustape": ("QN", "PI", "PB", "PE"),
    # "Dummy_Bare_bustape": ("", "PE"), # FIXME: missing dummy
    "Dummy_Pigtail_panel": ("QO", "PB"),
    "Dummy_PP0": ("QP", "PI"),
    "Dummy_Finger": ("QV", "PI"),
    "Dummy_Type_1_Data_link ": ("QQ", "PI", "PB", "PE"),
    "Dummy_Type_1_Power_DCS_line": ("QR", "PI", "PB", "PE"),
    "Dummy_Environmental_link": ("QS", "PI", "PB", "PE"),
    "Dummy_PP1_connector": ("QT", "PI", "PB", "PE"),
    "Dummy_PP1_connector_pieces_segments": ("QU", "PI", "PB", "PE"),
    # "Dummy_strain_relief": ("", "PB"), # FIXME: missing dummy
    "Dummy_Type_2_power_cable": ("QW", "PI", "PB", "PE"),
    "Dummy_Type_2_optobox_cable": ("QX", "PI", "PB", "PE"),
    "Dummy_PP2_box": ("QY", "PG"),
    "Dummy_Type_3_HV_cable": ("QZ", "PI", "PB", "PE"),
    "Dummy_Type_3_LV_cable": ("Q1", "PI", "PB", "PE"),
    "Dummy_MOPS_cable": ("Q2", "PG"),
    "Dummy_TiLock": ("Q3", "PG"),
    "Dummy_OPTO_MOPS": ("Q4", "PG"),
    "Dummy_PP3_power": ("Q5", "PG"),
}



from __future__ import annotations

import json
import pathlib

import pytest

import itksn

valid_sns = json.loads(
    (pathlib.Path(__file__).with_suffix("") / "valid_sns.json").read_text()
)


@pytest.mark.parametrize(
    "serial_number",
    valid_sns,
)
def test_sn(serial_number):
    itksn.parse(serial_number.encode("utf-8"))



from __future__ import annotations

import pytest
from construct import Bytes
from construct.core import MappingError, TerminatedError

import itksn
from itksn.core import EnumStr


def test_enumstr():
    myenum = EnumStr(Bytes(2), itsaa=b"aa", itsbb=b"bb", itsyy=b"yy")
    assert myenum.parse(b"yy").bytevalue == b"yy"
    assert myenum.parse(b"yy") == "itsyy"
    with pytest.raises(MappingError):
        myenum.parse(b"xx")


def test_parse_fe_wafer():
    parsed = itksn.parse(b"20UPGFW2123456")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "FE_chip_wafer"
    assert parsed.identifier.batch == "RD53A"
    assert parsed.identifier.number == b"2123456"


def test_parse_fe_wafer_wrong_subproject():
    with pytest.raises(MappingError):
        itksn.parse(b"20UPIFW2123456")


def test_parse_sensor():
    parsed = itksn.parse(b"20UPGW34212345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "Outer_pixel_sensor_wafer_150um_thickness"
    assert parsed.identifier.manufacturer == "V5_LFoundry"
    assert parsed.identifier.sensor_type == "Halfmoon_preproduction_Double_MS"
    assert parsed.identifier.number == b"12345"


def test_parse_bare_module():
    parsed = itksn.parse(b"20UPGBS1112345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "Digital_single_bare_module"
    assert parsed.identifier.FE_chip_version == "ITkpix_v1"
    assert parsed.identifier.sensor_type == "Market_survey_sensor_tile"
    assert parsed.identifier.number == b"12345"


def test_parse_pcb():
    parsed = itksn.parse(b"20UPGPD0112345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "Dual_PCB"
    assert parsed.identifier.FE_chip_version == "RD53A"
    assert parsed.identifier.PCB_manufacturer == "EPEC"
    assert parsed.identifier.number == b"12345"


def test_parse_module():
    parsed = itksn.parse(b"20UPGR90112345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "Digital_quad_module"
    assert parsed.identifier.FE_chip_version == "RD53A"
    assert parsed.identifier.PCB_manufacturer == "EPEC"
    assert parsed.identifier.number == b"12345"


def test_parse_quad_module():
    parsed = itksn.parse(b"20UPIM12602173")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == "L1_quad_module"
    assert parsed.identifier.FE_chip_version == "ITkpix_v1p1"
    assert parsed.identifier.PCB_manufacturer is None
    assert parsed.identifier.number == b"602173"


def test_parse_quad_module_rd53a():
    parsed = itksn.parse(b"20UPIM10602173")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == "L1_quad_module"
    assert parsed.identifier.FE_chip_version == "RD53A"
    assert parsed.identifier.PCB_manufacturer == "Yamashita_Material"
    assert parsed.identifier.number == b"02173"


def test_parse_module_carrier():
    parsed = itksn.parse(b"20UPGMC2291234")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "Module_carrier"
    assert parsed.identifier.module_type == "Linear_triplet_module_carrier"
    assert parsed.identifier.module_version == "Quad_v2p1"
    assert parsed.identifier.manufacturer == b"9"
    assert parsed.identifier.number == b"1234"


def test_parse_toomany():
    with pytest.raises(TerminatedError):
        itksn.parse(b"20UPGMC2291234999")


def test_parse_digital_quad_module():
    parsed = itksn.parse(b"20UPGR92101041")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "Digital_quad_module"
    assert parsed.identifier.FE_chip_version == "ITkpix_v1p1"
    assert parsed.identifier.PCB_manufacturer is None
    assert parsed.identifier.number == b"101041"


def test_parse_digital_quad_module_rd53a():
    parsed = itksn.parse(b"20UPGR90101041")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "Digital_quad_module"
    assert parsed.identifier.FE_chip_version == "RD53A"
    assert parsed.identifier.PCB_manufacturer == "EPEC"
    assert parsed.identifier.number == b"01041"


@pytest.mark.parametrize(
    ("serial_number", "component_code", "assembly_site", "number"),
    [
        ("20UPIMS2002140", "Triplet_L0_stave_module", "Genova", b"2140"),
        ("20UPIM52002140", "Triplet_L0_Ring0p5_module", "Genova", b"2140"),
        ("20UPIM02202123", "Triplet_L0_Ring0_module", "Oslo", b"2123"),
        ("20UPIMS2102148", "Triplet_L0_stave_module", "Barcelona", b"2148"),
    ],
    ids=["normal", "R05", "R0", "L0"],
)
def test_triplet_modules(serial_number, component_code, assembly_site, number):
    parsed = itksn.parse(serial_number.encode())
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == component_code
    assert parsed.identifier.assembly_site == assembly_site
    assert parsed.identifier.not_used == b"0"
    assert parsed.identifier.number == number


def test_fe_chip():
    parsed = itksn.parse(b"20UPGFC1048575")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_general"
    assert parsed.component_code == "FE_chip"
    assert parsed.identifier.batch == "ITkpix_v2"
    assert parsed.identifier.wafer == 255
    assert parsed.identifier.row == 15
    assert parsed.identifier.column == 15


def test_is_capillary():
    parsed = itksn.parse(b"20UPICP1299999")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == "IS_capillary"
    assert parsed.identifier.production_type == "Production"
    assert parsed.identifier.type == "R01"
    assert parsed.identifier.number == b"99999"


def test_outer_endcap_data_pp0():
    parsed = itksn.parse(b"20UPEDP2209999")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "pixel_endcaps"
    assert parsed.component_code == "Data_PP0"
    assert parsed.identifier.layer == "L2"
    assert parsed.identifier.flavor == "ring611_Front"
    assert parsed.identifier.number == b"9999"


def test_inner_system_pp1():
    parsed = itksn.parse(b"20UPI1P1000002")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == "PP1_connector"
    assert parsed.identifier.production_version == "Pre_production"
    assert parsed.identifier.glenair_part == "Type_2_Header"
    assert parsed.identifier.number == b"0002"


def test_inner_system_type1_power_L02xL1():
    parsed = itksn.parse(b"20UPIP19100001")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == "Type_1_Power_DCS_line"
    assert parsed.identifier.production_version == "Dummy"
    assert parsed.identifier.flavor == "L02xL1"
    assert parsed.identifier.number == b"00001"


def test_inner_system_type1_power_CR():
    parsed = itksn.parse(b"20UPIP10200001")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == "Type_1_Power_DCS_line"
    assert parsed.identifier.production_version == "Prototype"
    assert parsed.identifier.flavor == "Coupled_Ring"
    assert parsed.identifier.number == b"00001"


@pytest.mark.parametrize(
    ("serial_number", "region", "component_code", "component"),
    [
        ("20UPIPG0000000", "Barrel_Triplet", "Pigtail", "L0 Barrel Data Flex"),
        ("20UPIPP0000000", "Barrel_Triplet", "Power_pigtail", "L0 Barrel Power Flex"),
        ("20UPIPG0110000", "Barrel_Quad", "Pigtail", "L1 Barrel Data Flex F1"),
        ("20UPIPG0120000", "Barrel_Quad", "Pigtail", "L1 Barrel Data Flex F2"),
        ("20UPIPG0130000", "Barrel_Quad", "Pigtail", "L1 Barrel Data Flex F3"),
        ("20UPIPG0140000", "Barrel_Quad", "Pigtail", "L1 Barrel Data Flex F4"),
        ("20UPIPP0110000", "Barrel_Quad", "Power_pigtail", "L1 Barrel Power Flex F1"),
        ("20UPIPP0120000", "Barrel_Quad", "Power_pigtail", "L1 Barrel Power Flex F2"),
        ("20UPIRF0500000", "Endcap_Mixed", "Rigid_flex", "Coupled Ring"),
        ("20UPIRF0400000", "Endcap_Quad", "Rigid_flex", "Quad Ring"),
        ("20UPIRF0300000", "Endcap_Ring05_Triplet", "Rigid_flex", "Intermediate Ring"),
        ("20UPIPG0400000", "Endcap_Quad", "Pigtail", "Quad Module Z-Ray"),
        ("20UPIPG0210000", "Endcap_Ring0_Triplet", "Pigtail", "R0 Data Flex F1"),
        ("20UPIPG0220000", "Endcap_Ring0_Triplet", "Pigtail", "R0 Data Flex F2"),
        ("20UPIPG0230000", "Endcap_Ring0_Triplet", "Pigtail", "R0 Data Flex F3"),
        ("20UPIPP0510000", "Endcap_Mixed", "Power_pigtail", "Triplet Power T-Rex"),
        ("20UPIPP0520000", "Endcap_Mixed", "Power_pigtail", "Triplet Power Jumper"),
        ("20UPIPG0310000", "Endcap_Ring05_Triplet", "Pigtail", "R0.5 Data Flex F1"),
        ("20UPIPG0320000", "Endcap_Ring05_Triplet", "Pigtail", "R0.5 Data Flex F2"),
        ("20UPIPG0510000", "Endcap_Mixed", "Pigtail", "Type-0 to PP0 F1"),
        ("20UPIPG0520000", "Endcap_Mixed", "Pigtail", "Type-0 to PP0 F2"),
    ],
)
def test_inner_system_type0_services(serial_number, region, component_code, component):
    parsed = itksn.parse(serial_number.encode("utf-8"))
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel"
    assert parsed.subproject_code == "inner_pixel"
    assert parsed.component_code == component_code
    assert parsed.identifier.region == region
    assert parsed.identifier.number == b"0000"
    assert parsed.identifier.component == component



from __future__ import annotations

from construct import Bytes

from itksn.common import EnumStr

pcb_manufacturer = EnumStr(
    Bytes(1),
    Dummy=b"0",
    EPEC=b"1",
    NCAB_100um=b"2",
    ATLAFLEX=b"3",
    SFCircuits=b"4",
    PHOENIX=b"5",
    Yamashita_Material=b"6",
    NCAB_75um=b"7",
    Tecnomec=b"8",
)



from __future__ import annotations

from construct import (
    Bytes,
    Error,
    Switch,
)

from itksn.common import EnumStr
from itksn.pixels import local_supports, modules, services, utils

yy_identifiers = {
    **modules.yy_identifiers,
    **services.yy_identifiers,
    **local_supports.yy_identifiers,
}

subproject_codes = {
    subproject_code: EnumStr(
        Bytes(2),
        **{
            k: v.encode("utf-8")
            for k, (v, *allowed_subprojects) in yy_identifiers.items()
            if subproject_code in allowed_subprojects
        },
    )
    for subproject_code in ["PI", "PG", "PB", "PE"]
}

identifiers = Switch(
    lambda ctx: ctx.component_code,
    {
        "FE_chip_wafer": modules.fe_chip,
        "FE_chip": modules.fe_chip,
        "Planar_sensor_wafer_100um_thickness": modules.sensor,
        "Planar_sensor_wafer_150um_thickness": modules.sensor,
        "ThreeD_sensor_wafer": modules.sensor,
        "L0_inner_pixel_3D_sensor_wafer_25x100um": modules.sensor,
        "L0_inner_pixel_3D_sensor_wafer_50x50um": modules.sensor,
        "L1_inner_pixel_sensor_wafer_100um_thickness": modules.sensor,
        "Outer_pixel_sensor_wafer_150um_thickness": modules.sensor,
        "Half_size_planar_sensor_tile_100um_thickness": modules.sensor,
        "Half_size_planar_sensor_tile_150um_thickness": modules.sensor,
        "Full_size_planar_sensor_tile_100um_thickness": modules.sensor,
        "Full_size_planar_sensor_tile_150um_thickness": modules.sensor,
        "Half_size_3D_sensor_tile_25x100um": modules.sensor,
        "Full_size_3D_sensor_tile_25x100um": modules.sensor,
        "Half_size_3D_sensor_tile_50x50um": modules.sensor,
        "Full_size_3D_sensor_tile_50x50um": modules.sensor,
        "L0_inner_pixel_3D_sensor_tile_25x100um": modules.sensor,
        "L0_inner_pixel_3D_sensor_tile_50x50um": modules.sensor,
        "L1_inner_pixel_quad_sensor_tile": modules.sensor,
        "Outer_pixel_quad_sensor_tile": modules.sensor,
        "Planar_Sensor_test_structure_100um_thickness": modules.sensor,
        "Planar_Sensor_test_structure_150um_thickness": modules.sensor,
        "ThreeD_Sensor_test_structure_25x100um": modules.sensor,
        "ThreeD_Sensor_test_structure_50x50um": modules.sensor,
        "Planar_Sensor_half_moon_100um_thickness": modules.sensor,
        "Planar_Sensor_half_moon_150um_thickness": modules.sensor,
        "ThreeD_Sensor_half_moon_25x100um": modules.sensor,
        "ThreeD_Sensor_half_moon_50x50um": modules.sensor,
        "Single_bare_module": modules.bare_module,
        "Dual_bare_module": modules.bare_module,
        "Quad_bare_module": modules.bare_module,
        "Digital_single_bare_module": modules.bare_module,
        "Digital_quad_bare_module": modules.bare_module,
        "FourInch_bare_module_gel_pack": Error,
        "SixInch_bare_module_gel_pack": Error,
        "Triplet_L0_Stave_PCB": modules.pcb_triplets,
        "Triplet_L0_R0_PCB": modules.pcb_triplets,
        "Triplet_L0_R0p5_PCB": modules.pcb_triplets,
        "Quad_PCB": modules.pcb,
        "Dual_PCB": modules.pcb,
        "PCB_test_coupon": modules.pcb,
        "OB_wirebond_protection_roof": Error,
        "Triplet_L0_stave_module": modules.triplet_module,
        "Triplet_L0_Ring0_module": modules.triplet_module,
        "Triplet_L0_Ring0p5_module": modules.triplet_module,
        "L1_quad_module": modules.module,
        "Outer_system_quad_module": modules.module,
        "Dual_chip_module": modules.module,
        "Single_chip_module": modules.module,
        "Digital_triplet_L0_stave_module": modules.triplet_module,
        "Digital_triplet_L0_ring0_module": modules.triplet_module,
        "Digital_triplet_L0_ring0p5_module": modules.triplet_module,
        "Digital_quad_module": modules.module,
        "Digital_L1_quad_module": modules.module,
        "Dummy_triplet_L0_stave_module": modules.triplet_module,
        "Dummy_triplet_L0_ring0_module": modules.triplet_module,
        "Dummy_triplet_L0_ring0p5_module": modules.triplet_module,
        "Dummy_quad_module": modules.module,
        "Dummy_L1_quad_module": modules.module,
        "Module_carrier": modules.module_carrier,
        # dummy for testing of GUIs/tutorials
        "Dummy_FE_chip_wafer": modules.fe_chip,
        "Dummy_tutorial_FE_chip": modules.fe_chip,
        "Dummy_sensor_wafer": modules.sensor,
        "Dummy_tutorial_sensor_tile": modules.sensor,
        "Dummy_sensor_test_structure": modules.sensor,
        "Dummy_sensor_half_moon": modules.sensor,
        "Dummy_single_bare_module": modules.bare_module,
        "Dummy_quad_bare_module": modules.bare_module,
        "Dummy_tutorial_bare_module": modules.bare_module,
        "Dummy_bare_module_gel_pack": Error,
        "Dummy_tutorial_PCB": modules.pcb,
        "Dummy_PCB_test_coupon": modules.pcb,
        "Dummy_OB_wirebond_protection_roof": Error,
        "Dummy_tutorial_module": modules.module,
        "Dummy_module_carrier": modules.module_carrier,
        # local supports
        "IS_capillary": local_supports.local_supports_is,
        "IS_end_tube": local_supports.local_supports_is,
        "IS_cooling_tube": local_supports.local_supports_is,
        "IS_bare_local_support_stave": local_supports.local_supports_is,
        "IS_bare_local_support_ring": local_supports.local_supports_is,
        "IS_ring_local_support_assembly_loaded_ring": local_supports.local_supports,
        "IS_barrel_stave_assembly_loaded_stave": local_supports.local_supports,
        "OB_Base_Block": local_supports.local_supports,
        "OB_Cooling_Block": local_supports.local_supports,
        "OB_TPG_Tile": local_supports.local_supports,
        "OB_Local_Support_Inserts": local_supports.local_supports,
        "OB_Gusset": local_supports.local_supports,
        "OB_Truss": local_supports.local_supports,
        "OB_Half_Ring_Shell": local_supports.local_supports,
        "OB_End_of_longeron_Bracket_End_Gusset": local_supports.local_supports,
        "OB_Pipe_Support": local_supports.local_supports,
        "OB_Evaporator_Sleeves": local_supports.local_supports,
        "OB_Cooling_Pipe_IHR": local_supports.local_supports,
        "OB_Cooling_Pipe_Longeron": local_supports.local_supports,
        "OB_Functional_Pipe_for_IHR": local_supports.local_supports,
        "OB_Functional_Pipe_for_Longeron": local_supports.local_supports,
        "OB_End_of_longeron_Support": local_supports.local_supports,
        "OB_Bare_Module_Cell": local_supports.local_supports,
        "OB_Functional_IHR": local_supports.local_supports_ihr,
        "OB_Functional_Longeron": local_supports.local_supports_longeron,
        "OB_Loaded_Module_Cell": local_supports.loaded_local_supports_ob_module,
        "OB_Loaded_IHR": local_supports.local_supports,
        "OB_Loaded_Longeron": local_supports.local_supports,
        "OB_IHR_Handling_Frame": local_supports.local_supports,
        "OB_Longeron_Handling_Frame": local_supports.local_supports,
        "OB_Bare_Cell_Transport_Box": local_supports.local_supports,
        "OEC_Trapezoids": local_supports.local_supports,
        "OEC_Electrical break": local_supports.local_supports,
        "OEC_Inner_Rim_Insert": local_supports.local_supports,
        "OEC_Outer_Rim_mounting_lugs": local_supports.local_supports,
        "OEC_Inner_Rim_Closeout": local_supports.local_supports,
        "OEC_Outer_Rim_Closeout": local_supports.local_supports,
        "OEC_Pipe_Closeout_support_closeout": local_supports.local_supports,
        "OEC_Half_Sandwich": local_supports.local_supports,
        "OEC_Evaporator": local_supports.local_supports,
        "OEC_Bare_half_ring_assembly_Bare_support": local_supports.local_supports,
        "OEC_Loaded_local_support_loaded_support": local_supports.local_supports,
        "OEC_Handling_frame_support_frame": local_supports.local_supports,
        "OEC_Transport/storage_box_support_box": local_supports.local_supports,
        "Local_support_handling_frame_box": local_supports.local_supports_frame_box,
        "High_voltage_group": local_supports.local_supports,
        "Serial_powering_scheme": local_supports.local_supports,
        # dummy local supports
        "Dummy_IS_capillary": local_supports.local_supports_is,
        "Dummy_IS_end_tube": local_supports.local_supports_is,
        "Dummy_IS_cooling_tube": local_supports.local_supports_is,
        "Dummy_IS_bare_local_support_stave": local_supports.local_supports_is,
        "Dummy_IS_bare_local_support_ring": local_supports.local_supports_is,
        "Dummy_IS_ring_local_support_assembly_loaded_ring": local_supports.local_supports,
        "Dummy_IS_barrel_stave_assembly_loaded_stave": local_supports.local_supports,
        "Dummy_OB_Base_Block": local_supports.local_supports,
        "Dummy_OB_Cooling_Block": local_supports.local_supports,
        "Dummy_OB_TPG_Tile": local_supports.local_supports,
        "Dummy_OB_Local_Support_Inserts": local_supports.local_supports,
        "Dummy_OB_Gusset": local_supports.local_supports,
        "Dummy_OB_Truss": local_supports.local_supports,
        "Dummy_OB_Half_Ring_Shell": local_supports.local_supports,
        "Dummy_OB_End_of_longeron_Bracket_End_Gusset": local_supports.local_supports,
        "Dummy_OB_Pipe_Support": local_supports.local_supports,
        "Dummy_OB_Evaporator_Sleeves": local_supports.local_supports,
        "Dummy_OB_Cooling_Pipe_IHR": local_supports.local_supports,
        "Dummy_OB_Cooling_Pipe_Longeron": local_supports.local_supports,
        "Dummy_OB_Functional_Pipe_for_IHR": local_supports.local_supports,
        "Dummy_OB_Functional_Pipe_for_Longeron": local_supports.local_supports,
        "Dummy_OB_End_of_longeron_Support": local_supports.local_supports,
        "Dummy_OB_Bare_Module_Cell": local_supports.local_supports,
        "Dummy_OB_Functional_IHR": local_supports.local_supports_ihr,
        "Dummy_OB_Functional_Longeron": local_supports.local_supports_longeron,
        "Dummy_OB_Loaded_Module_Cell": local_supports.loaded_local_supports_ob_module,
        "Dummy_OB_Loaded_IHR": local_supports.local_supports,
        "Dummy_OB_Loaded_Longeron": local_supports.local_supports,
        "Dummy_OB_IHR_Handling_Frame": local_supports.local_supports,
        "Dummy_OB_Longeron_Handling_Frame": local_supports.local_supports,
        "Dummy_OB_Bare_Cell_Transport_Box": local_supports.local_supports,
        "Dummy_OEC_Trapezoids": local_supports.local_supports,
        "Dummy_OEC_Electrical break": local_supports.local_supports,
        "Dummy_OEC_Inner_Rim_Insert": local_supports.local_supports,
        "Dummy_OEC_Outer_Rim_mounting_lugs": local_supports.local_supports,
        "Dummy_OEC_Inner_Rim_Closeout": local_supports.local_supports,
        "Dummy_OEC_Outer_Rim_Closeout": local_supports.local_supports,
        "Dummy_OEC_Pipe_Closeout_support_closeout": local_supports.local_supports,
        "Dummy_OEC_Half_Sandwich": local_supports.local_supports,
        "Dummy_OEC_Evaporator": local_supports.local_supports,
        "Dummy_OEC_Bare_half_ring_assembly_Bare_support": local_supports.local_supports,
        "Dummy_OEC_Loaded_local_support_loaded_support": local_supports.local_supports,
        "Dummy_OEC_Handling_frame_support_frame": local_supports.local_supports,
        "Dummy_OEC_Transport/storage_box_support_box": local_supports.local_supports,
        "Dummy_Local_support_handling_frame_box": local_supports.local_supports_frame_box,
        "Dummy_High_voltage_group": local_supports.local_supports,
        "Dummy_Serial_powering_scheme": local_supports.local_supports,
        # services
        "Optoboard": services.optoboard,
        "Optoboard_termination_board": services.termination_board,  # OB Type-1 Termination Board (FIXME: only PG, no PB)
        "Optobox": services.optobox,
        "Optobox_powerbox": services.optobox,
        "Optobox_connector_board": services.optobox_powerboard_connector,
        "Optobox_optical_fan_out": services.optobox,
        "Optopanel": Error,  # FIXME
        "Optopanel_cooling_plate": Error,  # FIXME
        "Optobox_powerboard": services.optobox_powerboard_connector,
        "LpGBT_chip": Error,  # FIXME
        "GBCR_chip": Error,  # FIXME
        "Vtrx_module": Error,  # FIXME
        "Bpol2V5_chip": Error,  # FIXME
        "Bpol2V5_carrier_board": Error,  # FIXME
        "Bpol12V_chip": Error,  # FIXME
        "MOPS_chip": services.mops_chip,  # FIXME: phrasing in the document is awful
        "Power_cables": Error,  # FIXME
        "CAN_bus_cable": services.canbus,
        # Type-0 and Type-1 cables below
        "Pigtail": utils.subproject_switch(
            pi=services.is_type0_cable, pb=services.pb_type0_cable
        ),  # IS Type-0  (FIXME: no PI supported)
        "Rigid_flex": utils.subproject_switch(
            pi=services.is_type0_cable, pb=services.pb_type0_pp0
        ),  # IS Type-0
        "Data_PP0": utils.subproject_switch(
            pi=services.is_type0_cable, pe=services.pe_type0_data
        ),  # IS Type-0  # OB Type-1 Inclined PCB??? (FIXME: not PB)
        "Power_pigtail": utils.subproject_switch(
            pi=services.is_type0_cable, pe=services.pe_type0_power
        ),  # IS Type-0  # OE Type-0
        "Power_bustape": utils.subproject_switch(
            pe=services.pe_type0_power
        ),  # FIXME: only PE needed, not PI/PB
        "Bare_bustape": Error,  # FIXME: not used?
        "Pigtail_panel": services.pb_type0_cable,
        "PP0": utils.subproject_switch(pi=services.pi_type0_pp0),  # IS only
        "Finger": Error,  # FIXME: not used/defined?
        "Type_1_Data_link ": utils.subproject_switch(
            pb=services.pb_type1_data
        ),  # IS Type-1, (FIXME: only PB/PI needed, not PE)
        "Type_1_Power_DCS_line": utils.subproject_switch(
            pe=services.pe_type1,
            pb=services.pb_type1_power,
            pi=services.is_type1_cable,
        ),  # IS Type-1, (FIXME: only PB/PI needed, not PE)
        "Environmental_link": Error,  # Type-0 and Type-1 cable (FIXME: not defined?)
        "PP1_connector": utils.subproject_switch(
            pi=services.pp1, pe=services.pe_type1
        ),  # FIXME: not defined well)
        "PP1_connector_pieces_segments": utils.subproject_switch(),  # FIXME: not defined well)
        # End Type-0 and Type-1
        "Strain_relief": Error,
        "Type_2_power_cable": services.type2,
        "Type_2_optobox_cable": services.type2,
        "PP2_box": services.type2,
        "Type_3_HV_cable": services.type3,
        "Type_3_LV_cable": services.type3,
        "MOPS_cable": services.type3,
        "TiLock": services.type3,
        "OPTO_MOPS": services.type3,
        "PP3_power": services.type4,
        # Dummy services
        "Dummy_Optoboard": services.optoboard,
        "Dummy_Optoboard_termination_board": services.termination_board,
        "Dummy_Optobox": services.optobox,
        "Dummy_Optobox_powerbox": services.optobox,
        "Dummy_Optobox_connector_board": services.optobox_powerboard_connector,
        "Dummy_Optobox_optical_fan_out": services.optobox,
        "Dummy_Optopanel": Error,
        "Dummy_Optopanel_cooling_plate": Error,
        "Dummy_Optobox_powerboard": services.optobox_powerboard_connector,
        "Dummy_GBCR_chip": Error,
        "Dummy_Vtrx_module": Error,
        "Dummy_Bpol2V5_chip": Error,
        "Dummy_Bpol2V5_carrier_board": Error,
        "Dummy_Bpol12V_chip": Error,
        "Dummy_MOPS_chip": services.mops_chip,
        "Dummy_Power_cables": Error,
        "Dummy_CAN_bus_cable": services.canbus,
        "Dummy_Pigtail": utils.subproject_switch(
            pi=services.is_type0_cable, pb=services.pb_type0_cable
        ),  # IS Type-0
        "Dummy_Rigid_flex": utils.subproject_switch(
            pi=services.is_type0_cable, pb=services.pb_type0_pp0
        ),  # IS Type-0
        "Dummy_Data_PP0": utils.subproject_switch(
            pi=services.is_type0_cable, pe=services.pe_type0_data
        ),  # IS Type-0  # OB Type-1 Inclined PCB???
        "Dummy_Power_pigtail": utils.subproject_switch(
            pi=services.is_type0_cable, pe=services.pe_type0_power
        ),  # IS Type-0  # OE Type-0
        "Dummy_Power_bustape": utils.subproject_switch(pe=services.pe_type0_power),
        "Dummy_Pigtail_panel": services.pb_type0_cable,
        "Dummy_PP0": utils.subproject_switch(pi=services.pi_type0_pp0),  # IS only
        "Dummy_Finger": Error,
        "Dummy_Type_1_Data_link ": utils.subproject_switch(
            pb=services.pb_type1_data
        ),  # IS Type-1,
        "Dummy_Type_1_Power_DCS_line": utils.subproject_switch(
            pe=services.pe_type1, pb=services.pb_type1_power, pi=services.is_type1_cable
        ),  # IS Type-1,
        "Dummy_Environmental_link": Error,
        "Dummy_PP1_connector": utils.subproject_switch(
            pi=services.pp1, pe=services.pe_type1
        ),
        "Dummy_PP1_connector_pieces_segments": utils.subproject_switch(),
        "Dummy_Type_2_power_cable": services.type2,
        "Dummy_Type_2_optobox_cable": services.type2,
        "Dummy_PP2_box": services.type2,
        "Dummy_Type_3_HV_cable": services.type3,
        "Dummy_Type_3_LV_cable": services.type3,
        "Dummy_MOPS_cable": services.type3,
        "Dummy_TiLock": services.type3,
        "Dummy_OPTO_MOPS": services.type3,
        "Dummy_PP3_power": services.type4,
    },
    # default=Bytes(7),
)

__all__ = ("identifiers",)



from __future__ import annotations

from itksn import core
from itksn._version import __version__

parse = core.SerialNumberStruct.parse
build = core.SerialNumberStruct.build

__all__ = ["__version__", "build", "parse"]
del core



from __future__ import annotations

from construct import (
    Bytes,
    PaddedString,
    Struct,
    Switch,
    Terminated,
    this,
)

from itksn import pixels
from itksn.common import EnumStr

SerialNumberStruct = "SerialNumber" / Struct(
    "atlas_project" / EnumStr(Bytes(2), atlas_detector=b"20"),
    "system_code"
    / EnumStr(
        Bytes(1),
        phaseII_upgrade=b"U",
    ),
    "project_code"
    / EnumStr(
        Bytes(1),
        pixel=b"P",
        strip=b"S",
        common=b"C",
    ),
    "subproject_code"
    / Switch(
        lambda ctx: ctx.project_code,
        {
            "pixel": EnumStr(
                Bytes(1),
                inner_pixel=b"I",
                outer_pixel_barrel=b"B",
                pixel_general=b"G",
                pixel_endcaps=b"E",
            ),
            "strip": EnumStr(
                Bytes(1),
                strip_general=b"G",
                strip_barrel=b"B",
                strip_endcaps=b"E",
            ),
            "common": EnumStr(
                Bytes(1),
                common_mechanics=b"CM",
                common_electronics=b"CE",
            ),
        },
    ),
    "component_code"
    / Switch(
        this.subproject_code,
        {
            "inner_pixel": pixels.subproject_codes["PI"],
            "outer_pixel_barrel": pixels.subproject_codes["PB"],
            "pixel_general": pixels.subproject_codes["PG"],
            "pixel_endcaps": pixels.subproject_codes["PE"],
        },
        default=PaddedString(2, "utf8"),
    ),
    "identifier"
    / Switch(
        this.subproject_code,
        {
            "inner_pixel": pixels.identifiers,
            "outer_pixel_barrel": pixels.identifiers,
            "pixel_general": pixels.identifiers,
            "pixel_endcaps": pixels.identifiers,
        },
        default=Bytes(7),
    ),
    Terminated,
)
